# -*- coding: utf-8 -*-
import random
import warnings

import numpy as np

import scipy.spatial.distance as dist


def k_fold_cross_validation(X, K, randomise=False):
    """
    Generates K (training, validation) pairs from the items in X.

    Each pair is a partition of X, where validation is an iterable
    of length len(X)/K. So each training iterable is of length (K-1)*len(X)/K.

    If randomise is true, a copy of X is shuffled before partitioning,
    otherwise its order is preserved in training and validation.
    ## {{{ http://code.activestate.com/recipes/521906/ (r3)
    """
    if randomise:
        X = list(X)
        random.shuffle(X)
    for k in range(K):
        training = [x for i, x in enumerate(X) if i % K != k]
        validation = [x for i, x in enumerate(X) if i % K == k]
        yield training, validation


class GaussianProcess(object):
    """
    A simple class for Gaussian Process emulation. Currently, it assumes
    a squared exponential covariance function, but other covariance
    functions ought to be possible and easy to implement.

    """

    def __init__(self, inputs=None, targets=None, emulator_file=None):
        """The inputs are the input vectors, whereas the targets are the
        emulated model outputs.

        Parameters
        -----------
        inputs: array size (Ntrain x Ninputs)
                An input array of size Ntrain * Ninputs (Ntrain is the
                number of training samples, Ninputs is the dimensionality
                 of the input vector)
        targets: array size Ntrain
                The model outputs corresponding to the ``inputs`` training set
        """
        if emulator_file is None:
            if inputs is None:
                raise ValueError("No emulator file given, inputs can't be empty")
            if targets is None:
                raise ValueError("No emulator file given, targets can't be empty")
            self.inputs = inputs
            self.targets = targets
            (self.n, self.D) = self.inputs.shape

        else:
            if inputs is not None:
                raise ValueError("Can't have both emulator file and input")
            if targets is not None:
                raise ValueError("Can't have both emulator file and training")
            self._load_emulator(emulator_file)

    def _load_emulator(self, emulator_file):
        """Reads in emulator from npz file
        We read in an emulator parameterisation from an npz file.
        """
        emulator_f = np.load(emulator_file)
        self.emulator_file = emulator_file
        self.inputs = emulator_f['inputs']
        self.targets = emulator_f['targets']
        (self.n, self.D) = self.inputs.shape
        self.theta = emulator_f['theta']
        self._set_params(self.theta)

    def save_emulator(self, emulator_file):
        """Save emulator to disk as npz FileExistsError
        Saves an emulator to disk using an npz file.
        """
        np.savez(emulator_file, inputs=self.inputs, targets=self.targets,
                 theta=self.theta)

    def _prepare_likelihood(self):
        """
        This method precalculates matrices and stuff required for the i
        log-likelihood maximisation routine, so that they can be
        reused when calling the ``predict`` method repeatedly.
        """

        # Exponentiate the hyperparameters

        exp_theta = np.exp(self.theta)
        # Calculation of the covariance matrix Q using theta
        self.Z = dist.cdist(np.sqrt(exp_theta[: (self.D)])*self.inputs,
                            np.sqrt(exp_theta[: (self.D)])*self.inputs,
                            "sqeuclidean")
        self.Z = exp_theta[self.D] * np.exp(-0.5 * self.Z)
        self.Q = self.Z + exp_theta[self.D + 1] * np.eye(self.n)
        L = np.linalg.cholesky(self.Q)
        self.invQ = np.linalg.inv(L.T).dot(np.linalg.inv(L))
        # self.invQ = np.linalg.inv ( self.Q )
        self.invQt = np.dot(self.invQ, self.targets)

        self.logdetQ = 2.0 * np.sum(np.log(np.diag(L)))

    def loglikelihood(self, theta):
        """Calculates the loglikelihood for a set of hyperparameters
        ``theta``. The size of ``theta`` is given by the dimensions of
        the input vector to the model to be emulated.

        Parameters
        ----------
        theta: array
            Hyperparameters
        """
        self._set_params(theta)

        loglikelihood = (
            0.5 * self.logdetQ
            + 0.5 * np.dot(self.targets, self.invQt)
            + 0.5 * self.n * np.log(2. * np.pi)
        )
        self.current_theta = theta
        self.current_loglikelihood = loglikelihood
        return loglikelihood

    def partial_devs(self, theta):
        """This function calculates the partial derivatives of the
        cost function as a function of the hyperameters, and is only
        needed during GP training.

        Parameters
        -----------
        theta: array
                Hyperparameter set
        """
        partial_d = np.zeros(self.D + 2)

        for d in range(self.D):
            V = (
                (
                    (
                        np.tile(self.inputs[:, d], (self.n, 1))
                        - np.tile(self.inputs[:, d], (self.n, 1)).T
                    )
                )
                ** 2
            ).T * self.Z

            partial_d[d] = (
                np.exp(self.theta[d])
                * (np.dot(self.invQt, np.dot(V, self.invQt)) - np.sum(self.invQ * V))
                / 4.
            )

        partial_d[self.D] = 0.5 * np.sum(self.invQ * self.Z) - 0.5 * np.dot(
            self.invQt, np.dot(self.Z, self.invQt)
        )
        partial_d[self.D + 1] = 0.5 * np.trace(self.invQ) * np.exp(
            self.theta[self.D + 1]
        ) - 0.5 * np.dot(self.invQt, self.invQt) * np.exp(self.theta[self.D + 1])
        return partial_d

    def _set_params(self, theta):
        """Sets the hyperparameters, and thus also precalculates terms
        that depend on them. Since hyperparameters are fixed after
        training, this speeds up some calculations.

        Parameters
        -----------
        theta: array
                hyperparameters
        """

        self.theta = theta
        self._prepare_likelihood()

    def _learn(self, theta0, verbose):
        """The training method, called ''learn'' to keep up with the
        trendy Machine Learning kids!
        Takes an initial guess of the hyperparameters, and minimises
        that through a gradient descent algorithm, using methods
        ``likelihood`` and ``partial_devs`` to select hyperparameters
        that result in a minimal log-likelihood.

        Parameters
        -----------
        theta0: array
                Hyperparameters
        verbose: flag
                Whether to provide lots of information on the
                minimiation. Useful to see whether its fitting or
                not for some hairy problems.
        """
        # minimise self.loglikelihood (with self.partial_devs) to learn
        # theta
        from scipy.optimize import fmin_l_bfgs_b

        self._set_params(theta0)
        if verbose:
            iprint = 1
        else:
            iprint = -1
        try:
            # theta_opt = fmin_cg ( self.loglikelihood,
            #        theta0, fprime = self.partial_devs, \
            #        full_output=True, \
            #        retall = 1, disp=1 )
            theta_opt = fmin_l_bfgs_b(
                self.loglikelihood,
                theta0,
                fprime=self.partial_devs,
                factr=0.1,
                pgtol=1e-20,
                iprint=iprint,
            )
        except np.linalg.LinAlgError:
            warnings.warn(
                "Optimisation resulted in linear algebra error. " +
                "Returning last loglikelihood calculated, but this is fishy",
                RuntimeWarning,
            )
            # theta_opt = [ self.current_theta, self.current_loglikelihood ]
            theta_opt = [self.current_theta, 9999]

        return theta_opt

    def learn_hyperparameters(self, n_tries=15, verbose=False, x0=None):
        """User method to fit the hyperparameters of the model, using
        random initialisations of parameters. The user should provide
        a number of tries (e.g. how many random starting points to
        avoid local minima), and whether it wants lots of information
        to be reported back.

        Parameters
        -----------
        n_tries: int, optional
                Number of random starting points
        verbose: flag, optional
                How much information to parrot (e.g. convergence of
                the minimisation algorithm)
        x0: array, optional
                If you want to start the learning process with a
                particular vector, set it up here.

        """
        log_like = []
        params = []
        first = True
        for theta in 5. * (np.random.rand(n_tries, self.D + 2) - 0.5):
            if first and x0 is not None:
                first = False
                theta = x0
            T = self._learn(theta, verbose)
            log_like.append(T[1])
            params.append(T[0])
        log_like = np.array(log_like)
        idx = np.argsort(log_like)[0]
        print("After %d, the minimum cost was %e" % (n_tries, log_like[idx]))
        self._set_params(params[idx])
        return (log_like[idx], params[idx])

    def predict(self, testing, do_deriv=True, do_unc=True):
        """Make a prediction for a set of input vectors, as well as
        calculate the partial derivatives of the emulated model,
        and optionally, the "emulation uncertainty".

        Parameters
        -----------
        testing: array, size Npred * Ninputs
                The size of this array (and it must always be a 2D array!)
                is given by the number of input vectors that will be run
                through the emulator times the input vector size.

        do_unc: flag, optional
                Calculate the uncertainty (if you don't set this flag, it
                can shave a few us.
        do_deriv: flag, optional
                Whether to calculate the partial derivatives of the emulated
                model.

        Returns
        -------

        Three parameters (the mean, the variance and the partial derivatives)
        If some of those outputs have been left out, they are returned as
        `None` elements.
        """

        (nn, D) = testing.shape
        assert D == self.D

        expX = np.exp(self.theta)

        a = dist.cdist(
            np.sqrt(expX[: (self.D)]) * self.inputs,
            np.sqrt(expX[: (self.D)]) * testing,
            "sqeuclidean",
        )

        a = expX[self.D] * np.exp(-0.5 * a)
        b = expX[self.D]

        mu = np.dot(a.T, self.invQt)
        var = None
        deriv = None
        if do_unc:
            var = b - np.sum(a * np.dot(self.invQ, a), axis=0)
        # Derivative and partial derivatives of the function

        if do_deriv:
            deriv = np.zeros((nn, self.D))
            for d in range(self.D):
                aa = (
                    self.inputs[:, d].flatten()[None, :] -
                    testing[:, d].flatten()[:, None]
                )
                c = a * aa.T
                deriv[:, d] = expX[d] * np.dot(c.T, self.invQt)
        return mu, var, deriv

    def hessian(self, testing):
        """Calculates the hessian of the GP for the testing sample.
           hessian returns a (nn by d by d) array.
           
        """
        (nn, D) = testing.shape
        assert D == self.D
        expX = np.exp(self.theta)
        aprime = dist.cdist(
            np.sqrt(expX[: (self.D)]) * self.inputs,
            np.sqrt(expX[: (self.D)]) * testing,
            "sqeuclidean",
        )
        a = expX[self.D] * np.exp(-0.5 * aprime)
        dd_addition = np.identity(self.D) * expX[: (self.D)]
        hess = np.zeros((nn, self.D, self.D))
        for d in range(self.D):
            for d2 in range(self.D):
                aa = (
                    expX[d]
                    * (
                        self.inputs[:, d].flatten()[None, :]
                        - testing[:, d].flatten()[:, None]
                    )
                    * expX[d2]
                    * (
                        self.inputs[:, d2].flatten()[None, :]
                        - testing[:, d2].flatten()[:, None]
                    )
                    - dd_addition[d, d2]
                )
                cc = a * (aa.T)
                hess[:, d, d2] = np.dot(cc.T, self.invQt)
        return hess


if __name__ == "__main__":
    np.set_printoptions(precision=2, suppress=True)
    # input_obs = np.array ( [[-4, -3, -1, 0, 2]]).T
    # target1 = np.array ([-2, 0, 1., 2., -1])
    # data = np.loadtxt ("mvreg.dat", delimiter="," )
    # target1 = data[ :,0]
    # target2 = data[ :,1]
    # target3 = data[ :,2]
    # input_obs = data[ :, 3: ]
    wheat = np.loadtxt("argentine_wheat.dat")[:, 1:]
    yields = wheat[:, 0]
    mu = yields.mean()
    sigma = yields.std()
    yields = (yields - mu) / sigma
    wheat[:, 0] = yields
    rmse = []
    for (train, validate) in k_fold_cross_validation(wheat, 5, randomise=True):
        train = np.array(train)
        validate = np.array(validate)
        yields_t = train[:, 0]
        inputs_t = train[:, 1:]
        yields_v = validate[:, 0]
        inputs_v = validate[:, 1:]
        gp = GaussianProcess(inputs_t, yields_t)
        theta_min = gp.learn_hyperparameters(n_tries=2)
        pred_mu, pred_var, par_dev = gp.predict(inputs_v)
        print("TEST")
        print(inputs_v)
        print(pred_mu, pred_var, par_dev)
        r = (yields_v - pred_mu) ** 2  # /pred_var
        rmse.append([np.sqrt(r.mean()), theta_min[1]])

    # gp.theta[2] = 0.
    ######theta = gp.theta
    ######print theta
    ######gp.predict ( input_obs )

    ######x = np.linspace(-5,5,100)

    ######(mu, var ) = gp.predict ( x[:, np.newaxis])
    ######plt.plot ( x, mu, '-r' )
    ######plt.fill_between ( x, mu+np.sqrt(var), mu-np.sqrt(var), color='0.8')
    #######plt.errorbar ( x, mu, yerr=np.sqrt(var)*0.5 )
    ######plt.plot ( input_obs, target1, 'gs' )
    ######plt.title("theta: %s" % theta )
    ######plt.show()



# /usr/bin/env python
"""A multivariate Gaussian Process emulator through principal components compression.
This code provides a convenient way of emulating a multivariate model output (e.g. a
time series, or a spatial/temporal stack) by reducing it through principal component
analysis. Each of the selected PCs is then modelled through a Gaussian Process

Usage
-----

We need two arrays: one that stores the multivariate model output(`X`) and one that
stores the model parameters that produced said outputs (`y`). The sizes of these
arrays are important: `X` is `(N_training x N_full)` (i.e, `N_training` rows and
`N_full` elements in each column). `y` is `( N_train, N_params )`. Then simply

.. code-block:: python

   emulator = MultivariateEmulator( X=X, y=y )
   prediction, der_prediction = emulator.predict ( y[0] ) # Say

Saving emulators for future use
--------------------------------

If you want to save an emulator for re-use, you can do that easily by using the
`dump_emulator` method with a filename. You can then instantiate the
`MultivariateEmulator` class setting only the `dump` keyword in the constructor
to be the saved filename, and the emulator will be recreated.

"""

# import h5py

import numpy as np

from .GaussianProcess import GaussianProcess


class MultivariateEmulator(object):
    def __init__(
        self,
        dump=None,
        X=None,
        y=None,
        hyperparams=None,
        model="",
        sza=0,
        vza=0,
        raa=0,
        thresh=0.98,
        n_tries=5,
    ):
        """Constructor

        The constructor takes an array of model outputs `X`, and a vector
        of the parameters that served as the inputs for this model runs `y`.
        The sizes of these two arrays are `( N_training, N_full )` for `X` and
        `( N_train, N_params )` for `y`. `X` is decomposed into its PCs and
        the first `self.n_pcs` that explain `thresh` of the variance are
        selected. If `hyperparams` is set to `None`, normal training of the
        individual `self.n_pcs` GPs is carried out. If `hyperparams` is
        specified and it is the right shape `( N_params + 2, self.n_pcs )`,
        then there's no need for training the emulators (might be more
        efficient).

        Parameters
        ----------
        dump: str
            A filename that will be read to populate X, y and hyperparams
        X: array ( N_train, N_full )
            The modelled output array for training
        y: array (N_train, N_param )
            The corresponding training parameters for `X`
        hyperparams: array ( N_params + 2, N_PCs )
            The hyperparameters for the relevant GPs
        thresh: float
            The threshold at where to cutoff the percentage of
            variance explained.
        """
        if dump is not None:
            if X is None and y is None:
                if dump.find(".h5") > 0 or dump.find(".hdf5") > 0:
                    raise IOError("I can't be bothered working with HDF5 files")
                elif dump.find(".npz"):
                    self.emulator_file = dump
                    f = np.load(dump)
                    X = f["X"]
                    y = f["y"]
                    hyperparams = f["hyperparams"]
                    thresh = f["thresh"]
                    if "basis_functions" in dict(f):
                        basis_functions = f["basis_functions"]
                        n_pcs = f["n_pcs"]
                        f.close()
                else:
                    pass
            else:
                raise ValueError("You specified both a dump file and X and y")
        else:
            if X is None or y is None:
                raise ValueError("Need to specify both X and y")
            else:
                assert X.shape[0] == y.shape[0]
                assert X.ndim == 2
                assert y.ndim == 2
                basis_functions = None

        self.X_train = X
        self.y_train = y
        self.thresh = thresh
        if basis_functions is None:
            print("Decomposing the input dataset into basis functions...")
            self.calculate_decomposition(X, thresh)
            print("Done!\n ====> Using %d basis functions" % self.n_pcs)
            basis_functions = self.basis_functions
            n_pcs = self.n_pcs

        self.n_pcs = n_pcs
        self.basis_functions = basis_functions

        if hyperparams is not None:
            assert (y.shape[1] + 2 == hyperparams.shape[0]) and (
                self.n_pcs == hyperparams.shape[1]
            )
        self.train_emulators(X, y, hyperparams=hyperparams, n_tries=n_tries)

    def dump_emulator(self, fname, model_name, sza, vza, raa):
        """Save emulator to file for reuse

        Saves the emulator to a file (`.npz` format) for reuse.

        Parameters
        ----------
        fname: str
            The output filename

        """
        sza = int(sza)
        vza = int(vza)
        raa = int(raa)
        if fname.find(".npz") < 0 and (
            fname.find("h5") >= 0 or fname.find(".hdf") >= 0
        ):
            raise IOError("I can't be bothered working with HDF5 files")
        else:
            np.savez_compressed(
                fname,
                X=self.X_train,
                y=self.y_train,
                hyperparams=self.hyperparams,
                thresh=self.thresh,
                basis_functions=self.basis_functions,
                n_pcs=self.n_pcs,
            )
            print("Emulator safely saved")

    def calculate_decomposition(self, X, thresh):
        """Does PCA decomposition

        This simply does a PCA decomposition using the SVD. Note that
        if `X` is very large, more efficient methods of doing this
        might be required. The number of PCs to retain is selected
        as those required to estimate `thresh` of the total variance.

        Parameters
        -----------
        X: array ( N_train, N_full )
            The modelled output array for training
        thresh: float
            The threshold at where to cutoff the percentage of
            variance explained.
        """
        U, s, V = np.linalg.svd(X, full_matrices=False)
        pcnt_var_explained = s.cumsum() / s.sum()
        self.basis_functions = V[pcnt_var_explained <= thresh]
        self.n_pcs = np.sum(pcnt_var_explained <= thresh)

    def train_emulators(self, X, y, hyperparams, n_tries=2):
        """Train the emulators

        This sets up the required emulators. If necessary (`hypeparams`
        is set to None), it will train the emulators.

        X: array ( N_train, N_full )
            The modelled output array for training
        y: array (N_train, N_param )
            The corresponding training parameters for `X`
        hyperparams: array ( N_params + 2, N_PCs )
            The hyperparameters for the relevant GPs
        """
        self.emulators = []
        train_data = self.compress(X)
        self.hyperparams = np.zeros((2 + y.shape[1], self.n_pcs))
        for i in range(self.n_pcs):

            self.emulators.append(GaussianProcess(np.atleast_2d(y), train_data[i]))
            if hyperparams is None:
                print("\tFitting GP for basis function %d" % i)
                self.hyperparams[:, i] = self.emulators[i].learn_hyperparameters(
                    n_tries=n_tries
                )[1]
            else:
                self.hyperparams[:, i] = hyperparams[:, i]
                self.emulators[i]._set_params(hyperparams[:, i])

    def hessian(self, x):
        """A method to approximate the Hessian. This method builds on the fact
        that the spectral emulators are a linear combination of individual
        emulators. Therefore, we can calculate the Hessian of the spectral
        emulator as the sum of the individual products of individual Hessians
        times the spectral basis functions.
        """
        the_hessian = np.zeros((len(x), len(x), len(self.basis_functions[0])))
        x = np.atleast_2d(x)
        for i in range(self.n_pcs):
            # Calculate the Hessian of the weight
            this_hessian = self.emulators[i].hessian(x)
            # Add this hessian contribution once it's been scaled by the
            # relevant basis function.
            the_hessian = (
                the_hessian
                + this_hessian.squeeze()[:, :, None] * self.basis_functions[i]
            )
        return the_hessian

    def compress(self, X):
        """Project full-rank vector into PC basis"""
        return X.dot(self.basis_functions.T).T

    def predict(self, y, do_unc=True, do_deriv=True):
        """Prediction of input vector

        The individual GPs predict the PC weights, and these are used to
        reconstruct the value of the function at a point `y`. Additionally,
        the derivative of the function is also calculated. This is returned
        as a `( N_params, N_full )` vector (i.e., it needs to be reduced
        along axis 1)

        Parameters:
        y: array
            The value of the prediction point
        do_deriv: bool
            Whether derivatives are required or not
        do_unc: bool
            Whether to calculate the uncertainty or not

        Returns:
        A tuple with the predicted mean, predicted variance and
        patial derivatives. If any of the latter two elements have
        been switched off by `do_deriv` or `do_unc`, they'll be returned
        as `None`.
        """
        fwd = np.zeros(self.basis_functions[0].shape[0])
        y = np.atleast_2d(y)  # Just in case
        deriv = None
        unc = None
        if do_deriv:
            deriv = np.zeros((y.shape[1], self.basis_functions.shape[1]))
        if do_unc:
            unc = np.zeros_like(fwd)
        for i in range(self.n_pcs):
            pred_mu, pred_var, grad = self.emulators[i].predict(
                y, do_unc=do_unc, do_deriv=do_deriv
            )
            fwd += pred_mu * self.basis_functions[i]
            if do_deriv:
                deriv += (grad.squeeze()*self.basis_functions[i][:, None]).T
            if do_unc:
                unc += pred_var * self.basis_functions[i]
        try:
            return fwd.squeeze(), unc.squeeze(), deriv
        except AttributeError:
            return fwd.squeeze(), unc, deriv


###if __name__ == "__main__":
    #### read LUT to test this bits

    ###f = np.load("test_LUT.npz")
    ###angles = f["angles"]
    ###train_params = f["train_params"]
    ###train_brf = f["train_brf"]

    ###def unpack(params):
        ###"""Input a dictionary and output keys and array"""
        ###inputs = []
        ###keys = np.sort(list(params.keys()))
        ###for i, k in enumerate(keys):
            ###inputs.append(params[k])
        ###inputs = np.array(inputs).T
        ###return inputs, keys

    ###def pack(inputs, keys):
        ###"""Input keys and array and output dict"""
        ###params = {}
        ###for i, k in enumerate(keys):
            ###params[k] = inputs[i]
        ###return params

    ###train_paramsoot, keys = unpack(train_params.tolist())

    ###mv_em = MultivariateEmulator(X=train_brf, y=train_paramsoot)
    ###y_test = np.array([1., 0.5, 0.5, 1.65, 0.5, 0.5, 0.73, 0.89, 0.44, 0.42, 0.1])
    ###hypers = mv_em.hyperparams
    ###mv_em2 = MultivariateEmulator(X=train_brf, y=train_paramsoot, hyperparams=hypers)
    ###y_arr = y_test * 1
    ###for i in range(8):
        ###y_arr[-1] = 0.05 + 0.1 * i
        ###plt.plot(mv_em.predict(y_arr)[0], "-r", lw=2)
        ###plt.plot(mv_em2.predict(y_arr)[0], "-k", lw=1)
    ###mv_em.dump_emulator("emulator1.npz")
    ###plt.figure()
    ###new_em = MultivariateEmulator(dump="emulator1.npz")
    ###for i in range(8):
        ###y_arr[-1] = 0.05 + 0.1 * i
        ###plt.plot(mv_em.predict(y_arr)[0], "-r", lw=2)
        ###plt.plot(new_em.predict(y_arr)[0], "-k", lw=1)



#!/usr/bin/env python

__author__ = "J Gomez-Dans"
__copyright__ = "Copyright 2017, 2018 J Gomez-Dans"
__version__ = "1.6.10"
__license__ = "GPLv3"
__email__ = "j.gomez-dans@ucl.ac.uk"

from .GaussianProcess import GaussianProcess, k_fold_cross_validation
from .multivariate_gp import MultivariateEmulator
from .lhd import lhd
from .emulation_helpers import create_training_set, create_validation_set
from .emulation_helpers import create_emulator_validation
from .emulation_helpers import create_single_band_emulators
from .emulation_helpers import create_inverse_emulators



# -*- coding: utf-8 -*-
import multiprocessing
import numpy as np
import scipy.stats as ss
from .lhd import lhd
from .GaussianProcess import GaussianProcess
from .multivariate_gp import MultivariateEmulator


def integrate_passbands(spectrum, band_pass):
    """Integrate a spectrum over some band pass function.
    Parameters
    ----------
    spectrum: iter
        An array of spectra (n_spectra * n_wavelengths)
    band_pass: iter
        A 2D array of size `(n_bands x n_wavelengths)`. The sum over the
        second dimension of the array should evaluate to 1.

    Returns
    -------
    An array of n_spectra * n_bands with the integrated spectra.
    """

    n_bands = band_pass.shape[0]
    if band_pass.dtype == np.bool:
        x_pband = [spectrum[:, band_pass[i, :]].mean(axis=1) for i in range(n_bands)]
    else:
        band_pass = band_pass / (band_pass.sum(axis=1)[:, None])
        x_pband = [
            np.sum(spectrum[:, :] * band_pass[i, :], axis=1) for i in range(n_bands)
        ]

    x_train_pband = np.array(x_pband)
    return x_train_pband


def create_single_band_emulators(emulator, band_pass, n_tries=15):
    """This function creates per band emulators from the full-spectrum
    emulator. It requires passing an array of band pass functions of
    shape `(n_bands x n_wavelengths)` (e.g. (7 x 2101 for the 7 MODIS
    bands in the solar reflective domain 400-2500 nm with 1 nm spacing).
    The number of wavelengths should be compatible with that of the
    spectral emulator. Further, the sum of the band pass or spectral
    response function should be one (e.g. `np.sum(band_pass[i]) == 1`).

    Parameters
    ----------
    emulator: gp_emulator.MultivariateEmulator object
        A `gp_emulator.MultivariateEmulator` object that already exists.
    band_pass: iter
        A 2D array of size `(n_bands x n_wavelengths)`. The sum over the
        second dimension of the array should evaluate to 1.
    n_tries: int
        The number of random starts to train the single band emulator.
        Might be decreased for convenirnce to 5 or 10.

    Returns
    -------
    A list of `n_bands` `GaussianProcess` emulators.
    """

    x_train_pband = integrate_passbands(emulator.X_train, band_pass)
    emus = []
    for i in range(band_pass.shape[0]):
        gp = GaussianProcess(emulator.y_train[:] * 1.0, x_train_pband[i, :])
        gp.learn_hyperparameters(n_tries=n_tries)
        emus.append(gp)
    return emus


def create_inverse_emulators(original_emulator, band_pass, state_config, n_tries=10):
    """
    This function takes a multivariable output trained emulator
    and "retrains it" to take input reflectances and report a
    prediction of single input parameters (i.e., regression from
    reflectance/radiance to state). This is a useful starting
    point for spatial problems.

    Parameters
    ------------
    original_emulator: emulator
        An emulator (type gp_emulator.GaussianProcess)
    srf: array
        A 2d bandpass array (nbands, nfreq). Logical type
    state_config: ordered dict
        A state configuration ordered dictionary.
    """

    # For simplicity, let's get the training data out of the emulator
    X = original_emulator.X_train * 1.0
    y = original_emulator.y_train * 1.0
    # Apply band pass functions here...
    x_train_pband = integrate_passbands(X, band_pass)

    # A container to store the emulators
    gps = {}
    for i, param in enumerate(state_config):
        gp = GaussianProcess(x_train_pband.T, y[:, i])
        gp.learn_hyperparameters(n_tries=n_tries)
        gps[param] = gp
    return gps


def create_training_set(parameters, minvals, maxvals, fix_params=None, n_train=200):
    """Creates a traning set for a set of parameters specified by
    ``parameters`` (not actually used, but useful for debugging
    maybe). Parameters are assumed to be uniformly distributed
    between ``minvals`` and ``maxvals``. ``n_train`` input parameter
    sets will be produced, and returned with the actual distributions
    list. The latter is useful to create validation sets.

    It is often useful to add extra samples for regions which need to
    be carefully evaluated. We do this by adding a `fix_params` parameter
    which should be a dictionary indexing the parameter name, its fixed
    value, and the number of additional samples that will be drawn.

    Parameters
    -------------
    parameters: list
        A list of parameter names
    minvals: list
        The minimum value of the parameters. Same order as ``parameters``
    maxvals: list
        The maximum value of the parameters. Same order as ``parameters``
    fix_params: dictionary
        A diciontary indexed by the parameter name. Each item will have a
        tuple indicating the fixed value of the parameter, and how many
        extra LHS samples are required
    n_train: int
        How many training points to produce

    Returns
    ---------
    The training set and a distributions object that can be used by
    ``create_validation_set``
    """

    distributions = []
    for i, p in enumerate(parameters):
        distributions.append(
            ss.uniform(loc=minvals[i], scale=(maxvals[i] - minvals[i]))
        )
    samples = lhd(dist=distributions, size=n_train)

    if fix_params is not None:
        # Extra samples required
        for k, v in list(fix_params.items()):
            # Check whether they key makes sense
            if k not in parameters:
                raise ValueError(
                    "You have specified '%s', which is" % k
                    + " not in the parameters list"
                )

            extras = fix_parameter_training_set(
                parameters, minvals, maxvals, k, v[0], v[1]
            )
            samples = np.r_[samples, extras]

    return samples, distributions


def fix_parameter_training_set(
    parameters, minvals, maxvals, fixed_parameter, value, n_train
):
    """Produces a set of extra LHS samples where one parameter
    has been fixed to a single value, whereas all other parameters
    take their usual boundaries etc.

    Parameters
    ----------
    parameters: list
        List of parameter names
    minvals: array
        Minimum values of parameters. Order is same as in `parameters`
    maxvals: array
        Maximum values of parameters. Order is same as in `parameters`
    fixed_parameter: str
        The parameter that will be fixed.
    value: float
        The value of the the fixed parameter
    n_train: int
        Number of training samples

    Returns
    -------

    A new array of parameters
    """
    from copy import deepcopy  # groan

    parameters = deepcopy(parameters)
    minvals = deepcopy(minvals)
    maxvals = deepcopy(maxvals)
    fix_param = parameters.index(fixed_parameter)
    reduced_parameters = [p for p in parameters if p != fixed_parameter]
    minvals.pop(fix_param)
    maxvals.pop(fix_param)
    dummy_param = np.ones(n_train) * value
    distributions = []
    for i, p in enumerate(reduced_parameters):
        distributions.append(
            ss.uniform(loc=minvals[i], scale=(maxvals[i] - minvals[i]))
        )
    samples = lhd(dist=distributions, size=n_train)

    extra_array = np.insert(samples, fix_param, dummy_param, axis=1)
    return extra_array


def create_validation_set(distributions, n_validate=500):
    """Creates a validation set of ``n_validate`` vectors, using the
    ``distributions`` list.

    Parameters
    ------------
    distributions: list
        A set of parameter distributions (np.random objects)
    n_validate: int (optional)
        The number of samples to draw from `distributions`.

    Returns
    -------
    A bunch of samples
    """
    validate = []
    for d in distributions:
        validate.append(d.rvs(n_validate))
    validate = np.array(validate).T
    return validate


def create_emulator_validation(
    f_simulator,
    parameters,
    minvals,
    maxvals,
    n_train,
    n_validate,
    do_gradient=True,
    fix_params=None,
    thresh=0.98,
    n_tries=5,
    args=(),
    n_procs=None,
):
    """A method to create an emulator, given the simulator function, the
    parameters names and boundaries, the number of training input/output pairs.
    The function will also provide an independent validation dataset, both for
    the valuation of the function and its gradient. The gradient is calculated
    using finite differences, so it is a bit ropey.

    In order to better sample some regions of parameter space easily (you can
    change the underlying pdf of the parameters for LHS, but that's overkill)
    you can also add additional samples where one parameter is set to a fixed
    value, and an LHS design for all the other parameters is returned. This
    can be done usign the `fix_params` keyword.

    Parameters
    ------------
    f_simulator: function
        A function that evaluates the simulator. It should take a single
        parameter which will be made out of the input vector, plus whatever
        other extra arguments one needs (stored in ``args``).
    parameters: list
        The parameter names
    minvals: list
        The minimum value of the parameters
    maxvals: list
        The maximum value of the parameters
    n_train: int
        The number of training samples
    n_validate: int
        The number of validation samples
    thresh: float
        For a multivariate output GP, the threshold at which to cut the
        PCA expansion.
    n_tries: int
        The number of tries in the GP hyperparameter stage. The more the better,
        but also the longer it will take.
    args: tuple
        A list of extra arguments to the model
    do_gradient: Boolean
        Whether to do a gradient validation too.
    fix_params: dictionary
        A dictionary that allows the training set to be extended by fixing one
        or more parameters to one value, while still doing an LHS on the
        remaining parameters. Each parameter has a 2-element tuple, indicating
        the value and the number of extra samples.

    Returns
        The GP object, the validation input set, the validation output set, the
        emulated validation set, the emulated gradient set. If the gradient
        validation is also done, it will also return the gradient validation
        using finite differences.

    """

    # First, create the training set, using the appropriate function from
    # above...
    samples, distributions = create_training_set(
        parameters, minvals, maxvals, n_train=n_train, fix_params=fix_params
    )
    # Now, create the validation set, using the distributions object we got
    # from creating the training set
    validate = []
    for d in distributions:
        validate.append(d.rvs(n_validate))
    validate = np.array(validate).T

    # We have the input pairs for the training and validation. We will now run
    # the simulator function

    if n_procs is None:
        training_set = list(map(f_simulator, [((x,) + args) for x in samples]))
        validation_set = list(map(f_simulator, [((x,) + args) for x in validate]))

    else:
        pool = multiprocessing.Pool(processes=n_procs)

        training_set = pool.map(f_simulator, [((x,) + args) for x in samples])
        validation_set = pool.map(f_simulator, [((x,) + args) for x in validate])
    training_set = np.array(training_set).squeeze()
    validation_set = np.array(validation_set)

    if training_set.ndim == 1:
        gp = GaussianProcess(samples, training_set)
        gp.learn_hyperparameters(n_tries=n_tries)
    else:
        gp = MultivariateEmulator(
            X=training_set, y=samples, thresh=thresh, n_tries=n_tries
        )

    X = [gp.predict(np.atleast_2d(x)) for x in validate]
    if len(X[0]) == 2:
        emulated_validation = np.array([x[0] for x in X])
        emulated_gradient = np.array([x[1] for x in X])
    elif len(X[0]) == 3:
        emulated_validation = np.array([x[0] for x in X])
        emulated_gradient = np.array([x[2] for x in X])
    # Now with gradient... Approximate with finite differences...

    if do_gradient:
        val_set = [((x,) + args) for x in validate]
        validation_gradient = []
        delta = [(maxvals[j] - minvals[j]) / 10000.0 for j in range(len(parameters))]
        delta = np.array(delta)
        for i, pp in enumerate(val_set):
            xx0 = pp[0] * 1.0
            grad_val_set = []
            f0 = validation_set[i]
            df = []
            for j in range(len(parameters)):
                xx = xx0 * 1
                xx[j] = xx0[j] + delta[j]
                grad_val_set.append(xx)
                df.append(f_simulator(((xx,) + args)))
            df = np.array(df)
            try:
                validation_gradient.append((df - f0) / delta)
            except ValueError:
                validation_gradient.append((df - f0) / delta[:, None])

        return (
            gp,
            validate,
            validation_set,
            np.array(validation_gradient),
            emulated_validation,
            emulated_gradient.squeeze(),
        )
    else:
        return (gp, validate, validation_set, emulated_validation, emulated_gradient)



#!/usr/bin/env python
import codecs
import os
import re
from os import path
from setuptools import setup


this_directory = path.abspath(path.dirname(__file__))

with open(path.join(this_directory, 'README.md'), "rb") as f:
    long_description = f.read().decode('utf-8')

def read(*parts):
    with codecs.open(os.path.join(this_directory, *parts), 'r') as fp:
        return fp.read()
    
def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(name='gp_emulator',
      version=find_version("gp_emulator", "__init__.py"),
      description='A Python Gaussian Process emulator software package',
      classifiers=[
	'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Science/Research',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Environment :: Console'],
      author='J Gomez-Dans',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author_email='j.gomez-dans@ucl.ac.uk',
      url='http://github.com/jgomezdans/gp_emulator',
      packages=['gp_emulator'],
      install_requires=["numpy", "scipy"]
     )
